@@ -11,8 +11,10 @@
 //===----------------------------------------------------------------------===//
 
 #include "flang/Lower/SymbolMap.h"
+#include "flang/Optimizer/Builder/Todo.h"
 #include "mlir/IR/BuiltinTypes.h"
 #include "llvm/Support/Debug.h"
+#include <optional>
 
 #define DEBUG_TYPE "flang-lower-symbol-map"
 
@@ -25,11 +27,23 @@ void Fortran::lower::SymMap::addSymbol(Fortran::semantics::SymbolRef sym,
             [&](const fir::CharArrayBoxValue &v) { makeSym(sym, v, force); },
             [&](const fir::BoxValue &v) { makeSym(sym, v, force); },
             [&](const fir::MutableBoxValue &v) { makeSym(sym, v, force); },
+            [&](const fir::PolymorphicValue &v) { makeSym(sym, v, force); },
             [](auto) {
               llvm::report_fatal_error("value not added to symbol table");
             });
 }
 
+Fortran::lower::SymbolBox toSymbolBox(
+    std::variant<Fortran::lower::SymbolBox, fir::FortranVariableOpInterface>
+        symboxOrdefiningOp) {
+  if (const Fortran::lower::SymbolBox *symBox =
+          std::get_if<Fortran::lower::SymbolBox>(&symboxOrdefiningOp))
+    return *symBox;
+  auto definingOp =
+      std::get<fir::FortranVariableOpInterface>(symboxOrdefiningOp);
+  TODO(definingOp.getLoc(), "FortranVariableOpInterface lookup as SymbolBox");
+}
+
 Fortran::lower::SymbolBox
 Fortran::lower::SymMap::lookupSymbol(Fortran::semantics::SymbolRef symRef) {
   Fortran::semantics::SymbolRef sym = symRef.get().GetUltimate();
@@ -37,7 +51,7 @@ Fortran::lower::SymMap::lookupSymbol(Fortran::semantics::SymbolRef symRef) {
        jmap != jend; ++jmap) {
     auto iter = jmap->find(&*sym);
     if (iter != jmap->end())
-      return iter->second;
+      return toSymbolBox(iter->second);
   }
   return SymbolBox::None{};
 }
@@ -47,7 +61,7 @@ Fortran::lower::SymbolBox Fortran::lower::SymMap::shallowLookupSymbol(
   auto &map = symbolMapStack.back();
   auto iter = map.find(&symRef.get().GetUltimate());
   if (iter != map.end())
-    return iter->second;
+    return toSymbolBox(iter->second);
   return SymbolBox::None{};
 }
 
@@ -65,7 +79,7 @@ Fortran::lower::SymbolBox Fortran::lower::SymMap::lookupOneLevelUpSymbol(
   for (++jmap; jmap != jend; ++jmap) {
     auto iter = jmap->find(&*sym);
     if (iter != jmap->end())
-      return iter->second;
+      return toSymbolBox(iter->second);
   }
   return SymbolBox::None{};
 }
@@ -78,6 +92,23 @@ Fortran::lower::SymMap::lookupImpliedDo(Fortran::lower::SymMap::AcDoVar var) {
   return {};
 }
 
+std::optional<fir::FortranVariableOpInterface>
+Fortran::lower::SymMap::lookupVariableDefinition(semantics::SymbolRef symRef) {
+  Fortran::semantics::SymbolRef sym = symRef.get().GetUltimate();
+  for (auto jmap = symbolMapStack.rbegin(), jend = symbolMapStack.rend();
+       jmap != jend; ++jmap) {
+    auto iter = jmap->find(&*sym);
+    if (iter != jmap->end()) {
+      if (const auto *varDef =
+              std::get_if<fir::FortranVariableOpInterface>(&iter->second))
+        return *varDef;
+      else
+        return std::nullopt;
+    }
+  }
+  return std::nullopt;
+}
+
 llvm::raw_ostream &
 Fortran::lower::operator<<(llvm::raw_ostream &os,
                            const Fortran::lower::SymbolBox &symBox) {
@@ -92,15 +123,29 @@ Fortran::lower::operator<<(llvm::raw_ostream &os,
   return os;
 }
 
+static llvm::raw_ostream &
+dump(llvm::raw_ostream &os,
+     const std::variant<Fortran::lower::SymbolBox,
+                        fir::FortranVariableOpInterface> &symboxOrdefiningOp) {
+  if (const Fortran::lower::SymbolBox *symBox =
+          std::get_if<Fortran::lower::SymbolBox>(&symboxOrdefiningOp))
+    return os << *symBox;
+  auto definingOp =
+      std::get<fir::FortranVariableOpInterface>(symboxOrdefiningOp);
+  return os << definingOp << "\n";
+}
+
 llvm::raw_ostream &
 Fortran::lower::operator<<(llvm::raw_ostream &os,
                            const Fortran::lower::SymMap &symMap) {
   os << "Symbol map:\n";
   for (auto i : llvm::enumerate(symMap.symbolMapStack)) {
     os << " level " << i.index() << "<{\n";
-    for (auto iter : i.value())
+    for (auto iter : i.value()) {
       os << "  symbol @" << static_cast<const void *>(iter.first) << " ["
-         << *iter.first << "] ->\n    " << iter.second;
+         << *iter.first << "] ->\n    ";
+      dump(os, iter.second);
+    }
     os << " }>\n";
   }
   return os;
