@@ -9,7 +9,7 @@
 // <complex>
 
 // complex& operator=(const complex&);
-// template<class X> complex& operator= (const complex<X>&);
+// template<class X> complex& operator= (const complex<X>&); // constexpr in C++20
 
 #include <complex>
 #include <cassert>
@@ -17,7 +17,8 @@
 #include "test_macros.h"
 
 template <class T, class X>
-void
+TEST_CONSTEXPR_CXX20
+bool
 test()
 {
     std::complex<T> c;
@@ -31,6 +32,7 @@ test()
     c = c3;
     assert(c.real() == 3.5);
     assert(c.imag() == -4.5);
+    return true;
 }
 
 int main(int, char**)
@@ -47,5 +49,19 @@ int main(int, char**)
     test<long double, double>();
     test<long double, long double>();
 
+#if TEST_STD_VER >= 20
+    static_assert(test<float, float>());
+    static_assert(test<float, double>());
+    static_assert(test<float, long double>());
+
+    static_assert(test<double, float>());
+    static_assert(test<double, double>());
+    static_assert(test<double, long double>());
+
+    static_assert(test<long double, float>());
+    static_assert(test<long double, double>());
+    static_assert(test<long double, long double>());
+#endif
+
   return 0;
 }
